import httpx
import pytest

from kiarina.lib.firebase.auth import (
    FirebaseAPIError,
    InvalidRefreshTokenError,
    exchange_custom_token,
    refresh_id_token,
    settings_manager,
)


async def test_invalid_refresh_token(load_settings) -> None:
    settings = settings_manager.get_settings()

    with pytest.raises(InvalidRefreshTokenError):
        await refresh_id_token(
            refresh_token="invalid_refresh_token",
            api_key=settings.api_key.get_secret_value(),
        )


async def test_firebase_api_error() -> None:
    with pytest.raises(FirebaseAPIError):
        await refresh_id_token(
            refresh_token="invalid_refresh_token",
            api_key="invalid_api_key",
        )


async def test_httpx_reqest_error(monkeypatch) -> None:
    def mock_post(*args, **kwargs):
        raise httpx.RequestError("Network error")

    monkeypatch.setattr("httpx.AsyncClient.post", mock_post)

    with pytest.raises(FirebaseAPIError) as exc_info:
        await refresh_id_token(
            refresh_token="some_refresh_token",
            api_key="invalid_api_key",
        )

    assert "Request failed" in str(exc_info.value)


async def test_happy_path(firebase_app) -> None:
    auth = pytest.importorskip("firebase_admin.auth")

    custom_token = auth.create_custom_token("test").decode("utf-8")

    settings = settings_manager.get_settings()
    exchange_response = await exchange_custom_token(
        custom_token=custom_token,
        api_key=settings.api_key.get_secret_value(),
    )

    response = await refresh_id_token(
        refresh_token=exchange_response.refresh_token,
        api_key=settings.api_key.get_secret_value(),
    )

    assert response.id_token
    assert response.refresh_token



import httpx

from .._exceptions.firebase_api_error import FirebaseAPIError
from .._exceptions.invalid_refresh_token_error import InvalidRefreshTokenError
from .._schemas.token_response import TokenResponse


async def refresh_id_token(refresh_token: str, api_key: str) -> TokenResponse:
    """
    Refresh ID token using refresh token.

    Raises:
        InvalidRefreshTokenError: If refresh token is invalid or expired
        FirebaseAPIError: If Firebase API returns an error
    """
    url = f"https://securetoken.googleapis.com/v1/token?key={api_key}"

    payload = {
        "grant_type": "refresh_token",
        "refresh_token": refresh_token,
    }

    async with httpx.AsyncClient() as client:
        try:
            response = await client.post(url, json=payload, timeout=30.0)
            response.raise_for_status()

        except httpx.HTTPStatusError as e:
            error_data = e.response.json() if e.response.content else {}
            error_message = error_data.get("error", {}).get("message", str(e))
            error_code = error_data.get("error", {}).get("code")

            # Check for invalid refresh token error
            if (
                "INVALID_REFRESH_TOKEN" in error_message
                or "TOKEN_EXPIRED" in error_message
            ):
                raise InvalidRefreshTokenError(
                    f"Invalid or expired refresh token: {error_message}"
                ) from e

            raise FirebaseAPIError(
                message=f"Firebase API error: {error_message}",
                status_code=e.response.status_code,
                error_code=error_code,
            ) from e

        except httpx.RequestError as e:
            raise FirebaseAPIError(f"Request failed: {e}") from e

    data = response.json()

    return TokenResponse(
        id_token=data["id_token"],
        refresh_token=data["refresh_token"],
        expires_in=int(data["expires_in"]),
    )



import os

import pytest
from pydantic_settings_manager import load_user_configs

import kiarina.utils.file as kf


@pytest.fixture(scope="session")
def load_settings():
    env_var = "KIARINA_LIB_FIREBASE_AUTH_TEST_SETTINGS_FILE"

    if env_var not in os.environ:
        pytest.skip(f"Environment variable {env_var} not set, skipping tests.")

    test_settings_file = os.environ[env_var]
    test_settings_file = os.path.expanduser(test_settings_file)

    if not os.path.exists(test_settings_file):
        raise FileNotFoundError(f"Settings file not found: {test_settings_file}")

    user_configs = kf.read_yaml_dict(test_settings_file)

    if not user_configs:
        raise ValueError(f"Settings file is empty or invalid: {test_settings_file}")

    load_user_configs(user_configs)


@pytest.fixture(scope="session")
def firebase_app(load_settings):
    """Initialize Firebase Admin SDK once per test session."""
    firebase_admin = pytest.importorskip("firebase_admin")
    credentials = pytest.importorskip("firebase_admin.credentials")

    import kiarina.lib.google.auth

    service_account_file = (
        kiarina.lib.google.auth.settings_manager.get_settings().service_account_file
    )
    assert service_account_file is not None

    # Initialize Firebase Admin SDK (only once per session)
    app = firebase_admin.initialize_app(credentials.Certificate(service_account_file))

    yield app

    # Cleanup after all tests
    firebase_admin.delete_app(app)



import httpx

from .._exceptions.firebase_api_error import FirebaseAPIError
from .._exceptions.invalid_custom_token_error import InvalidCustomTokenError
from .._schemas.token_response import TokenResponse


async def exchange_custom_token(custom_token: str, api_key: str) -> TokenResponse:
    """
    Exchange a Firebase custom token for an ID token.

    Raises:
        InvalidCustomTokenError: If the custom token is invalid
        FirebaseAPIError: If Firebase API returns an error
    """
    url = f"https://identitytoolkit.googleapis.com/v1/accounts:signInWithCustomToken?key={api_key}"

    payload = {
        "token": custom_token,
        "returnSecureToken": True,
    }

    async with httpx.AsyncClient() as client:
        try:
            response = await client.post(url, json=payload, timeout=30.0)
            response.raise_for_status()

        except httpx.HTTPStatusError as e:
            error_data = e.response.json() if e.response.content else {}
            error_message = error_data.get("error", {}).get("message", str(e))
            error_code = error_data.get("error", {}).get("code")

            # Check for invalid custom token error
            if "INVALID_CUSTOM_TOKEN" in error_message:
                raise InvalidCustomTokenError(
                    f"Invalid custom token: {error_message}"
                ) from e

            raise FirebaseAPIError(
                message=f"Firebase API error: {error_message}",
                status_code=e.response.status_code,
                error_code=error_code,
            ) from e

        except httpx.RequestError as e:
            raise FirebaseAPIError(f"Request failed: {e}") from e

    data = response.json()

    return TokenResponse(
        id_token=data["idToken"],
        refresh_token=data["refreshToken"],
        expires_in=int(data["expiresIn"]),
    )



import asyncio
from datetime import datetime, timedelta, timezone

from .._utils.refresh_id_token import refresh_id_token
from .._schemas.token_response import TokenResponse


class TokenManager:
    def __init__(
        self,
        refresh_token: str,
        api_key: str,
        *,
        id_token: str | None = None,
        expires_at: datetime | None = None,
        refresh_buffer_seconds: int = 300,
    ):
        self.refresh_token: str = refresh_token
        self.api_key: str = api_key
        self._id_token: str | None = id_token
        self._expires_at: datetime | None = expires_at
        self._refresh_buffer_seconds = refresh_buffer_seconds
        self._refresh_lock = asyncio.Lock()

    @property
    def id_token(self) -> str:
        if self._id_token is None:  # pragma: no cover
            raise AssertionError("ID token is not set.")

        return self._id_token

    @property
    def expires_at(self) -> datetime:
        if self._expires_at is None:  # pragma: no cover
            raise AssertionError("Expiration time is not set.")

        return self._expires_at

    async def get_id_token(self) -> str:
        if self._needs_refresh():
            async with self._refresh_lock:
                # Double-check after acquiring lock
                if self._needs_refresh():
                    await self._do_refresh()

        assert self._id_token is not None
        return self._id_token

    async def refresh(self) -> TokenResponse:
        async with self._refresh_lock:
            return await self._do_refresh()

    def _needs_refresh(self) -> bool:
        if self._id_token is None or self._expires_at is None:
            return True

        now = datetime.now(timezone.utc)
        refresh_threshold = self._expires_at - timedelta(
            seconds=self._refresh_buffer_seconds
        )

        return now >= refresh_threshold

    async def _do_refresh(self) -> TokenResponse:
        response = await refresh_id_token(self.refresh_token, self.api_key)

        self.refresh_token = response.refresh_token
        self._id_token = response.id_token
        self._expires_at = response.expires_at

        return response



from .firebase_auth_error import FirebaseAuthError


class FirebaseAPIError(FirebaseAuthError):
    """Firebase API returned an error response."""

    def __init__(
        self,
        message: str,
        status_code: int | None = None,
        error_code: str | None = None,
    ):
        super().__init__(message)

        self.status_code = status_code
        self.error_code = error_code



import asyncio

import pytest

from kiarina.lib.firebase.auth import (
    TokenManager,
    exchange_custom_token,
    settings_manager,
)


async def test_token_manager(firebase_app):
    auth = pytest.importorskip("firebase_admin.auth")
    custom_token = auth.create_custom_token("test").decode("utf-8")

    settings = settings_manager.get_settings()
    api_key = settings.api_key.get_secret_value()

    response = await exchange_custom_token(
        custom_token=custom_token,
        api_key=api_key,
    )

    manager = TokenManager(
        refresh_token=response.refresh_token,
        api_key=api_key,
    )

    id_token = await manager.get_id_token()
    expires_at = manager.expires_at
    assert id_token == manager.id_token

    id_token_2 = await manager.get_id_token()
    assert id_token_2 == id_token
    assert manager.expires_at == expires_at

    await asyncio.sleep(0.1)

    await manager.refresh()
    assert manager.expires_at > expires_at



from pydantic import SecretStr
from pydantic_settings import BaseSettings, SettingsConfigDict
from pydantic_settings_manager import SettingsManager


class FirebaseAuthSettings(BaseSettings):
    model_config = SettingsConfigDict(
        env_prefix="KIARINA_LIB_FIREBASE_AUTH_",
        extra="ignore",
    )

    project_id: str
    """
    Firebase project ID
    """

    api_key: SecretStr
    """
    Firebase Web API Key

    Obtain this key from the Firebase Console.
    Project Settings > General > Your apps > Web API Key
    """


settings_manager = SettingsManager(FirebaseAuthSettings, multi=True)



import logging
from importlib import import_module
from typing import TYPE_CHECKING

if TYPE_CHECKING:  # pragma: no cover
    from ._exceptions.firebase_api_error import FirebaseAPIError
    from ._exceptions.firebase_auth_error import FirebaseAuthError
    from ._exceptions.invalid_custom_token_error import InvalidCustomTokenError
    from ._exceptions.invalid_refresh_token_error import InvalidRefreshTokenError
    from ._schemas.token_response import TokenResponse
    from ._services.token_manager import TokenManager
    from ._settings import FirebaseAuthSettings, settings_manager
    from ._utils.exchange_custom_token import exchange_custom_token
    from ._utils.refresh_id_token import refresh_id_token

__all__ = [
    # ._exceptions
    "FirebaseAPIError",
    "FirebaseAuthError",
    "InvalidCustomTokenError",
    "InvalidRefreshTokenError",
    # ._schemas
    "TokenResponse",
    # ._services
    "TokenManager",
    # ._settings
    "FirebaseAuthSettings",
    "settings_manager",
    # ._utils
    "exchange_custom_token",
    "refresh_id_token",
]

logging.getLogger(__name__).addHandler(logging.NullHandler())


def __getattr__(name: str) -> object:
    if name not in __all__:  # pragma: no cover
        raise AttributeError(f"module {__name__} has no attribute {name}")

    module_map = {
        # ._exceptions
        "FirebaseAPIError": "._exceptions.firebase_api_error",
        "FirebaseAuthError": "._exceptions.firebase_auth_error",
        "InvalidCustomTokenError": "._exceptions.invalid_custom_token_error",
        "InvalidRefreshTokenError": "._exceptions.invalid_refresh_token_error",
        # ._schemas
        "TokenResponse": "._schemas.token_response",
        # ._services
        "TokenManager": "._services.token_manager",
        # ._settings
        "FirebaseAuthSettings": "._settings",
        "settings_manager": "._settings",
        # ._utils
        "exchange_custom_token": "._utils.exchange_custom_token",
        "refresh_id_token": "._utils.refresh_id_token",
    }

    globals()[name] = getattr(import_module(module_map[name], __name__), name)
    return globals()[name]



from datetime import datetime, timedelta, timezone

from pydantic import BaseModel


class TokenResponse(BaseModel):
    id_token: str
    """Firebase ID token (JWT)."""

    refresh_token: str
    """Refresh token for getting new ID tokens."""

    expires_in: int
    """ID token lifetime in seconds."""

    @property
    def expires_at(self) -> datetime:
        """
        Calculate expiration datetime (UTC).
        """
        return datetime.now(timezone.utc) + timedelta(seconds=self.expires_in)
